/*
 * error_management1.sql
 * Chapter 3, Oracle10g PL/SQL Programming
 * by Ron Hardman, Michael McLaughlin and Scott Urman
 *
 * This script builds a stored procedure to manage and format
 * errors for the error stack.
 */

-- Unremark for debugging script.
-- SET ECHO ON
SET SERVEROUTPUT ON SIZE 1000000

-- Build an anonymous block that will trigger an error.
CREATE OR REPLACE PROCEDURE handle_errors
( object_name           IN        VARCHAR2
, module_name           IN        VARCHAR2 := NULL
, table_name            IN        VARCHAR2 := NULL
, sql_error_code        IN        NUMBER   := NULL
, sql_error_message     IN        VARCHAR2 := NULL
, user_error_message    IN        VARCHAR2 := NULL ) IS

  -- Define local exceptioon variable.
  raised_error          EXCEPTION;

  -- Define a nested table to manage exceptions.
  TYPE error_stack IS TABLE OF VARCHAR2(200);
  
  -- Define a local nested table variable.
  errors                ERROR_STACK := error_stack();
  
  FUNCTION object_type
  ( object_name_in      IN        VARCHAR2 )
  RETURN VARCHAR2 IS
  
    -- Define default type.
    return_type         VARCHAR2(12) := 'Unidentified';
  
  BEGIN
  
    FOR i IN ( SELECT    object_type
               FROM     user_objects
               WHERE    object_name = object_name_in ) LOOP
               
      return_type := i.object_type;
      
    END LOOP;
    
    -- Return discovered or unknown type.
    RETURN return_type;

  END object_type;

BEGIN

    -- Extend and append to error stack.
  errors.EXTEND;
  errors(errors.COUNT) := object_type(object_name)||' ['||object_name||']';
  
  -- Check if module name not null and append to error.
  IF module_name IS NOT NULL THEN
  
    -- Extend and append to error stack.
    errors.EXTEND;
    errors(errors.COUNT) := 'Module Name: ['||module_name||']';
  
  END IF;
  
  -- Check if table name not null and append to error.
  IF table_name IS NOT NULL THEN

    -- Extend and append to error stack.
    errors.EXTEND;
    errors(errors.COUNT) := 'Table Name: ['||table_name||']';
   
  END IF;
  
  -- Check if SQLCODE not null and append to error.
  IF sql_error_code IS NOT NULL THEN
  
    -- Extend and append to error stack.
    errors.EXTEND;
    errors(errors.COUNT) := 'SQLCODE Value: ['||sql_error_code||']';
  
  END IF;
  
  -- Check if SQLERRM not null and append to error.
  IF sql_error_message IS NOT NULL THEN
  
    -- Extend and append to error stack.
    errors.EXTEND;
    errors(errors.COUNT) := 'SQLERRM Value: ['||sql_error_message||']';
  
  END IF;
  
  -- Check if user error message not null and append to error.
  IF user_error_message IS NOT NULL THEN
  
    -- Extend and append to error stack.
    errors.EXTEND;
    errors(errors.COUNT) := 'User Error Message: ['||module_name||']';
  
  END IF;

  -- Print a line to break the error stack.
  errors.EXTEND;
  errors(errors.COUNT) := '----------------------------------------';
  
  -- Raise an exception.
  RAISE raised_error;

EXCEPTION
  
  WHEN raised_error THEN
    FOR i IN 1..errors.COUNT LOOP
      dbms_output.put_line(errors(i));
    END LOOP;
    RETURN;

END;
/

show errors

-- Drop any objects to make script re-runnable.
BEGIN

  FOR i IN (SELECT   table_name
            FROM     user_tables
            WHERE    table_name = 'DUMMY' ) LOOP
            
    -- Use NDS to drop the dependent object type.
    EXECUTE IMMEDIATE 'DROP TABLE dummy';
    
  END LOOP;
  
END;
/

-- Create a dummy table.
CREATE TABLE dummy
( dummy_value       VARCHAR2(1));


-- Create a third level procedure.
CREATE OR REPLACE PROCEDURE error_level3 IS

  -- Define local variables for error stack management.
  local_object            VARCHAR2(30)  := 'ERROR_LEVEL3';
  local_module            VARCHAR2(30)  := 'MAIN';
  local_table             VARCHAR2(30)  :=  NULL;
  local_user_message      VARCHAR2(200) :=  NULL;
  
BEGIN

  -- Insert a value to large for the column.
  INSERT INTO dummy VALUES (11);
  
EXCEPTION

  -- Raise exception by calling the handle_errors procedure.
  WHEN others THEN
    handle_errors( object_name => local_object
                 , module_name => local_module
                 , sql_error_code => SQLCODE
                 , sql_error_message => SQLERRM
                 , user_error_message => DBMS_UTILITY.FORMAT_ERROR_BACKTRACE );
    RAISE;

END error_level3;
/

show errors

CREATE OR REPLACE PROCEDURE error_level2 IS

  -- Define local variables for error stack management.
  local_object            VARCHAR2(30)  := 'ERROR_LEVEL2';
  local_module            VARCHAR2(30)  := 'MAIN';
  local_table             VARCHAR2(30)  :=  NULL;
  local_user_message      VARCHAR2(200) :=  NULL;
  
BEGIN

  -- Call next level.
  error_level3();

EXCEPTION

  -- Raise exception by calling the handle_errors procedure.
  WHEN others THEN
    handle_errors( object_name => local_object
                 , module_name => local_module
                 , sql_error_code => SQLCODE
                 , sql_error_message => SQLERRM
                 , user_error_message => DBMS_UTILITY.FORMAT_ERROR_BACKTRACE );
    RAISE;

END error_level2;
/

show errors

CREATE OR REPLACE PROCEDURE error_level1 IS

  -- Define local variables for error stack management.
  local_object            VARCHAR2(30)  := 'ERROR_LEVEL1';
  local_module            VARCHAR2(30)  := 'MAIN';
  local_table             VARCHAR2(30)  :=  NULL;
  local_user_message      VARCHAR2(200) :=  NULL;
  
BEGIN

  -- Call next level.
  error_level2();

EXCEPTION

  -- Raise exception by calling the handle_errors procedure.
  WHEN others THEN
    handle_errors( object_name => local_object
                 , module_name => local_module
                 , sql_error_code => SQLCODE
                 , sql_error_message => SQLERRM
                 , user_error_message => DBMS_UTILITY.FORMAT_ERROR_BACKTRACE );
    RAISE;

END error_level1;
/

show errors

-- Define an anonomyous test block.
BEGIN

  -- Start the chain of calls.
  error_level1;
  
END;
/