/*
 * create_invoker1.sql
 * Chapter 4, Oracle Expert PL/SQL Programming
 * by Ron Hardman and Michael McLaughlin
 *
 * This script builds a definer's-rights package and grants
 * EXECUTE privileges to the MYAPP user.
 */

-- Unremark for debugging script.
SET ECHO ON
SET SERVEROUTPUT ON SIZE 1000000

-- Drop table if it exists.
BEGIN

  -- Only drop table if it is there.
  FOR i IN (SELECT   object_name
            FROM     user_objects
            WHERE    object_name = 'LOCAL_TABLE' ) LOOP
  
    -- Drop the table.
    EXECUTE IMMEDIATE 'DROP TABLE '||i.object_name;
    
  END LOOP;
  
END;
/

-- Create a local table.
CREATE TABLE local_table
( owner_name VARCHAR2(30 CHAR)
, user_name  VARCHAR2(30 CHAR) );

-- Create an INVOKER_RIGHTS package.
CREATE OR REPLACE PACKAGE invoker_rights AUTHID CURRENT_USER IS

  -- Define a SET_DATA function.
  FUNCTION set_data
  ( user_in VARCHAR2 )
  RETURN BOOLEAN;
  
  -- Define a GET_DATA function.
  FUNCTION get_data
  ( user_in VARCHAR2 )
  RETURN VARCHAR2;
  
END;
/

SHOW ERRORS

-- Create an INVOKER_RIGHTS package body.
CREATE OR REPLACE PACKAGE BODY invoker_rights IS

  -- Define a SET_DATA function.
  FUNCTION set_data
  ( user_in VARCHAR2 )
  RETURN BOOLEAN AS
  
    -- Define default return value.
    retval BOOLEAN := FALSE;
  
  BEGIN
  
    -- Use an implicit for-loop to avoid declaring variables.
    INSERT
    INTO     local_table
    VALUES
    ('PLSQL'
    , user_in );
    
    COMMIT;
    
    -- Reset return value.
    retval := TRUE;
    
    -- Return value.
    RETURN retval;

  END set_data;
  
  -- Define a GET_DATA function.
  FUNCTION get_data
  ( user_in VARCHAR2 )
  RETURN VARCHAR2 AS
  
  BEGIN
  
    -- Use an implicit for-loop to avoid declaring variables.
    FOR i IN (SELECT   user_in
              ,        owner_name
              ,        user_name
              FROM     local_table
              WHERE    user_name = user_in ) LOOP
              
      -- Return the value.
      RETURN 'Called by ['||user_in||']['||i.owner_name||']['||i.user_name||']';
      
    END LOOP;

  END get_data;
  
END invoker_rights;
/

SHOW ERRORS

-- Test the definer table from the definer schema.
BEGIN

  -- Call the function that inserts the user name.
  IF invoker_rights.set_data(USER) THEN

    -- Use an implicit for-loop to avoid declaring variables.
    FOR i IN (SELECT   invoker_rights.get_data(user) data

              FROM     dual ) loop
              
      -- Print the value returned.
      DBMS_OUTPUT.PUT_LINE(i.data);
      
    END LOOP;
    
  END IF;
  
END;
/

-- Clean out the table.
TRUNCATE TABLE local_table;

-- Grant EXECUTE privileges to MYAPP schema.
GRANT EXECUTE ON invoker_rights TO myapp;
