/*
 * create_telephone_objects.sql
 * Chapter 6, Oracle Expert PL/SQL
 * by Ron Hardman and Michael McLaughlin
 *
 * This script builds the data tables for Chapter 6 and seeds
 * them with data.
 */

CLEAR BREAKS
CLEAR COLUMNS
CLEAR COMPUTES

TTITLE OFF

SPOOL create_telephone_objects.log

-- Run dependent classes.
@cleanup.sql

SET ECHO ON
SET FEEDBACK ON
SET NULL '<Null>'
SET PAGESIZE 999
SET PAUSE OFF
SET TERM ON
SET TIME OFF
SET TIMING OFF
SET VERIFY OFF

CREATE OR REPLACE TYPE telephone_id_table
  IS TABLE OF NUMBER;
/

-- Create the telephone object.
CREATE OR REPLACE TYPE telephone_object
AS OBJECT
( telephone_id         NUMBER
, contact_id           NUMBER
, address_id           NUMBER
, telephone_type       NUMBER
, country_code         VARCHAR2(3 CHAR)
, area_code            VARCHAR2(6 CHAR)
, telephone_number     VARCHAR2(10 CHAR)
, CONSTRUCTOR FUNCTION telephone_object
  ( telephone_id         NUMBER
  , contact_id           NUMBER
  , address_id           NUMBER
  , telephone_type       NUMBER
  , country_code         VARCHAR2
  , area_code            VARCHAR2
  , telephone_number     VARCHAR2 )
  RETURN SELF AS RESULT 
, CONSTRUCTOR FUNCTION telephone_object
  ( telephone_id         NUMBER   )
  RETURN SELF AS RESULT
, MEMBER FUNCTION get_telephone_number
  ( telephone_id         NUMBER )
  RETURN VARCHAR2
, MEMBER PROCEDURE delete_telephone_object
, STATIC PROCEDURE delete_telephone_object
  ( telephone_id         NUMBER )
, MEMBER PROCEDURE insert_telephone_object
, STATIC PROCEDURE insert_telephone_object
  ( telephone_id         NUMBER
  , contact_id           NUMBER
  , address_id           NUMBER
  , telephone_type       NUMBER
  , country_code         VARCHAR2
  , area_code            VARCHAR2
  , telephone_number     VARCHAR2 )
, MEMBER PROCEDURE update_telephone_object
, STATIC PROCEDURE update_telephone_object
  ( telephone_id         NUMBER
  , country_code         VARCHAR2
  , area_code            VARCHAR2
  , telephone_number     VARCHAR2 )
, MEMBER PROCEDURE print_telephone_number
, MEMBER PROCEDURE print_telephone_object
, ORDER MEMBER FUNCTION equals
  ( class_in TELEPHONE_OBJECT )
  RETURN NUMBER )
INSTANTIABLE NOT FINAL;
/

SHOW ERRORS

-- Create the telephone object body.
CREATE OR REPLACE TYPE BODY telephone_object AS

  -- ------------------------------------------------------------

  -- Define a complete attribute constructor.
  CONSTRUCTOR FUNCTION telephone_object
  ( telephone_id       NUMBER
  , contact_id         NUMBER
  , address_id         NUMBER
  , telephone_type     NUMBER
  , country_code       VARCHAR2
  , area_code          VARCHAR2
  , telephone_number   VARCHAR2 )
  RETURN SELF AS RESULT IS

  BEGIN

    -- Map and assign the column values to object attributes.
    SELF.telephone_id := telephone_id;
    SELF.contact_id := contact_id;
    SELF.address_id := address_id;
    SELF.telephone_type := telephone_type;
    SELF.country_code := country_code;
    SELF.area_code := area_code;
    SELF.telephone_number := telephone_number;

    -- Return instance of the object type.
    RETURN;

  END;

  -- ------------------------------------------------------------

  -- Define a unique key attribute constructor.
  CONSTRUCTOR FUNCTION telephone_object
  ( telephone_id         NUMBER )
  RETURN SELF AS RESULT IS

    -- Define a cursor to get set elements.
    CURSOR get_telephone
    ( telephone_id_in IN NUMBER ) IS
      SELECT   *
      FROM     telephone
      WHERE    telephone_id = telephone_id_in;

  BEGIN

    -- Use a for-loop to traverse the cursor.
    FOR i IN get_telephone(telephone_id) LOOP

      -- Map and assign the column values to object attributes.
      SELF.telephone_id := i.telephone_id;
      SELF.contact_id := i.contact_id;
      SELF.address_id := i.address_id;
      SELF.telephone_type := i.telephone_type;
      SELF.country_code := i.country_code;
      SELF.area_code := i.area_code;
      SELF.telephone_number := i.telephone_number;

    END LOOP;

    -- Return instance of the object type.
    RETURN;

  END;

  -- ------------------------------------------------------------

  -- Define a function to build a telphone number.
  MEMBER FUNCTION get_telephone_number
  ( telephone_id         NUMBER ) RETURN VARCHAR2 IS

    -- Define return variable.
    retval     VARCHAR2(22 CHAR);

    -- Define a cursor to get telephone number.
    CURSOR get_telephone
    ( telephone_id_in IN NUMBER ) IS
      SELECT   country_code
      ,        area_code
      ,        telephone_number
      FROM     telephone
      WHERE    telephone_id = telephone_id_in;

  BEGIN

    -- Query using a FOR-LOOP to ensure a null is returned when 
    -- no row is found.
    FOR i IN get_telephone(telephone_id) LOOP 

      retval :=
        '[('||i.area_code||') '||i.telephone_number||']';

    END LOOP;

    -- Solely to illustrate a null print return.
    IF retval IS NULL THEN

      retval := '[]';

    END IF;

    -- Return variable length string.
    RETURN retval;

  END;

  -- ------------------------------------------------------------

  -- Define a procedure to delete a telephone object.
  MEMBER PROCEDURE delete_telephone_object IS

    -- Define variable for DML statement.
    statement VARCHAR2(4000 CHAR);

  BEGIN

    -- Build DML statement for Native Dynamic SQL (NDS).
    statement := 'DELETE '
              || 'FROM     telephone '
              || 'WHERE    telephone_id = '''
              ||  SELF.telephone_id||'''';

    -- Use Native Dynamic SQL to execute DML statement.
    EXECUTE IMMEDIATE statement;

  END;

  -- ------------------------------------------------------------

  -- Define a procedure to delete a telephone object.
  STATIC PROCEDURE delete_telephone_object
  ( telephone_id         NUMBER ) IS

    -- Define variable for DML statement.
    statement VARCHAR2(4000 CHAR);

  BEGIN

    -- Build DML statement for Native Dynamic SQL (NDS).
    statement := 'DELETE '
              || 'FROM     telephone '
              || 'WHERE    telephone_id = '''
              ||  telephone_id||'''';

    -- Use Native Dynamic SQL to execute DML statement.
    EXECUTE IMMEDIATE statement;

  END;

  -- ------------------------------------------------------------

  -- Define a procedure to insert a telephone object.
  MEMBER PROCEDURE insert_telephone_object IS

  BEGIN

    -- Use a static statement.
    INSERT
    INTO     telephone
    VALUES
    ( SELF.telephone_id
    , SELF.contact_id
    , SELF.address_id
    , SELF.telephone_type
    , SELF.country_code
    , SELF.area_code
    , SELF.telephone_number);

  END;

  -- ------------------------------------------------------------

  -- Define a static procedure to insert a telephone object.
  STATIC PROCEDURE insert_telephone_object
  ( telephone_id         NUMBER
  , contact_id           NUMBER
  , address_id           NUMBER
  , telephone_type       NUMBER
  , country_code         VARCHAR2
  , area_code            VARCHAR2
  , telephone_number     VARCHAR2 ) IS

  BEGIN

    -- Use a static statement.
    INSERT
    INTO     telephone
    VALUES
    ( telephone_id
    , contact_id
    , address_id
    , telephone_type
    , country_code
    , area_code
    , telephone_number);

  END;

  -- ------------------------------------------------------------

  -- Define a procedure to update a telephone object.
  MEMBER PROCEDURE update_telephone_object IS

    -- Define variable for DML statement.
    statement VARCHAR2(4000 CHAR);

  BEGIN

    -- Build DML statement for Native Dynamic SQL (NDS).
    statement := 'UPDATE   telephone '
              || 'SET      country_code = '''
              ||             SELF.country_code||''''
              || ',        area_code = '''
              ||             SELF.area_code||''''
              || ',        telephone_number = '''
              ||             SELF.telephone_number||''''
              || 'WHERE    telephone_id = '
              ||             SELF.telephone_id;

    -- Use Native Dynamic SQL (NDS) to execute DML statement.
    EXECUTE IMMEDIATE statement;

    COMMIT;

  END;

  -- ------------------------------------------------------------

  -- Define a static procedure to update a telephone object.
  STATIC PROCEDURE update_telephone_object
  ( telephone_id         NUMBER
  , country_code         VARCHAR2
  , area_code            VARCHAR2
  , telephone_number     VARCHAR2 ) IS

    -- Define variable for DML statement.
    statement VARCHAR2(4000 CHAR);

  BEGIN

    -- Build DML statement for Native Dynamic SQL (NDS).
    statement := 'UPDATE   telephone '
              || 'SET      country_code = '''
              ||             country_code||''''
              || ',        area_code = '''
              ||             area_code||''''
              || ',        telephone_number = '''
              ||             telephone_number||''''
              || 'WHERE    telephone_id = '
              ||             telephone_id;

    -- Use Native Dynamic SQL (NDS) to execute DML statement.
    EXECUTE IMMEDIATE statement;

    COMMIT;

  END;

  -- ------------------------------------------------------------

  -- Member procedure to print the formatted object.
  MEMBER PROCEDURE print_telephone_number IS

  BEGIN

    -- Print formatted telephone number.
    DBMS_OUTPUT.PUT_LINE('Telephone '
    ||                   '[('||SELF.area_code||') '
    ||                   SELF.telephone_number||']');

  END;

  -- ------------------------------------------------------------

  -- Member procedure to print the elements of the object.
  MEMBER PROCEDURE print_telephone_object IS

  BEGIN

    -- Print object attributes.
    DBMS_OUTPUT.PUT_LINE('SELF.TELEPHONE_ID     ['
    ||                    SELF.telephone_id||']');
    DBMS_OUTPUT.PUT_LINE('SELF.CONTACT_ID       ['
    ||                    SELF.contact_id||']');
    DBMS_OUTPUT.PUT_LINE('SELF.ADDRESS_ID       ['
    ||                    SELF.address_id||']');
    DBMS_OUTPUT.PUT_LINE('SELF.TELEPHONE_TYPE   ['
    ||                    SELF.telephone_type||']');
    DBMS_OUTPUT.PUT_LINE('SELF.COUNTRY_CODE     ['
    ||                    SELF.country_code||']');
    DBMS_OUTPUT.PUT_LINE('SELF.AREA_CODE        ['
    ||                    SELF.area_code||']');
    DBMS_OUTPUT.PUT_LINE('SELF.TELEPHONE_NUMBER ['
    ||                    SELF.telephone_number||']');

  END;

  -- ------------------------------------------------------------

  -- A method to check for equality of object instances.
  ORDER MEMBER FUNCTION equals
  ( class_in TELEPHONE_OBJECT )
  RETURN NUMBER IS

    -- Define a default return values.
    false_value NUMBER := 0;
    true_value  NUMBER := 1;

  BEGIN

    -- Compare all elements in the object.
    IF SELF.telephone_id = class_in.telephone_id AND
      SELF.contact_id = class_in.contact_id AND
      SELF.address_id = class_in.address_id AND
      SELF.telephone_type = class_in.telephone_type AND
      SELF.country_code = class_in.country_code AND
      SELF.area_code = class_in.area_code AND
      SELF.telephone_number = class_in.telephone_number THEN

      -- Return True they are equal.
      RETURN true_value;

    ELSE 

      -- Return False they are unequal.
      RETURN false_value;

    END IF;
      
  END;

  -- ------------------------------------------------------------

END;
/ 

SHOW ERRORS

-- Create a table of telephone objects.
CREATE OR REPLACE TYPE telephone_object_table
AS TABLE OF telephone_object;
/

SHOW ERRORS

-- Create the set of telephone objects.
CREATE OR REPLACE TYPE telephone_object_set
AS OBJECT
  ( telephone_entries    TELEPHONE_OBJECT_TABLE
  , set_size             NUMBER
, CONSTRUCTOR FUNCTION telephone_object_set
  ( telephone_entries    TELEPHONE_OBJECT_TABLE )
  RETURN SELF AS RESULT 
, CONSTRUCTOR FUNCTION telephone_object_set
  ( contact_id           NUMBER   )
  RETURN SELF AS RESULT
, MEMBER PROCEDURE add_set_member
  ( telephone          TELEPHONE_OBJECT )
, MEMBER FUNCTION get_collection
  RETURN TELEPHONE_OBJECT_TABLE
, MEMBER FUNCTION get_set_member
  ( telephone_id         NUMBER   )
  RETURN TELEPHONE_OBJECT
, MEMBER PROCEDURE delete_telephone_object_set
  ( contact_id         NUMBER )
, MEMBER PROCEDURE insert_telephone_object_set
  ( telephone_table      TELEPHONE_OBJECT_TABLE )
, MEMBER PROCEDURE update_telephone_object_set
  ( telephone_entries    TELEPHONE_OBJECT_TABLE )
, MEMBER PROCEDURE print_telephone_number_set
, MEMBER PROCEDURE print_telephone_object_set
, ORDER MEMBER FUNCTION equals
  ( class_in TELEPHONE_OBJECT_SET )
  RETURN NUMBER )
INSTANTIABLE NOT FINAL;
/

SHOW ERRORS

CREATE OR REPLACE TYPE BODY telephone_object_set AS

  -- ------------------------------------------------------------

  -- Construct using a table collection.
  CONSTRUCTOR FUNCTION telephone_object_set
  ( telephone_entries  TELEPHONE_OBJECT_TABLE )
  RETURN SELF AS RESULT IS

  BEGIN

    -- Construct an instance of the telephone_object_set.
    SELF.telephone_entries := telephone_entries;

    -- Assign the size value to an instance variable.
    SELF.set_size := telephone_entries.LAST;

    -- Return an instance of the telephone_object_set.
    RETURN;

  END;

  -- ------------------------------------------------------------

  -- Construct using a subtype key in a table.
  CONSTRUCTOR FUNCTION telephone_object_set
  ( contact_id         NUMBER )
  RETURN SELF AS RESULT IS

    -- Define a local counter variable.
    c PLS_INTEGER := 1;

    -- Define a object structure.
    telephone TELEPHONE_OBJECT;
    
    -- Define and initialize the object collection.
    telephones TELEPHONE_OBJECT_TABLE := telephone_object_table(); 

    -- Define a cursor to get set elements.
    CURSOR get_telephone
    ( contact_id_in IN NUMBER) IS
      SELECT   telephone_id
      FROM     telephone
      WHERE    contact_id = contact_id_in;

  BEGIN

    -- Use a for-loop to traverse the cursor.
    FOR i IN get_telephone(contact_id) LOOP

      -- Construct an instance of telephone_object.
      telephone := telephone_object(i.telephone_id);

      -- Allot space to the collection.
      telephones.EXTEND;

      -- Assign an element to the collection.
      -- --------------------------------------------------------
      --  The index value for the collection may be a PLS_INTEGER
      --  but cannot be the for-loop counter variable "i". If you
      --  attempt to use the for-loop counter variable, you will
      --  raise an ORA-00382 error, which is an "expression of 
      --  wrong type" message.
      -- --------------------------------------------------------
      telephones(c) := telephone;

      -- Increment the counter.
      c := c + 1;

    END LOOP;

    -- Instantiate the object with the table collection.
    SELF.telephone_entries := telephones;

    -- Assign the size value to an instance variable.
    SELF.set_size := telephones.LAST;

    -- Return the object instance.
    RETURN;

  END;

  -- ------------------------------------------------------------

  -- Add a member to the object collection of the instance.
  MEMBER PROCEDURE add_set_member
  ( telephone            TELEPHONE_OBJECT ) IS

  BEGIN

    -- Extend the local collection size. 
    SELF.telephone_entries.EXTEND;

    -- Assign a new element.
    -- ---------------------------------------------
    --   When the Oracle 10g Collection API EXTEND
    --   method is called, the size of the table
    --   collection is incremented by one.
    -- ---------------------------------------------
    SELF.telephone_entries(SELF.telephone_entries.LAST) := telephone;

    -- Increment the known size of the collection.
    SELF.set_size := SELF.set_size + 1;

  END;

  -- ------------------------------------------------------------

  -- Function to convert the object to a collection.
  MEMBER FUNCTION get_collection
  RETURN TELEPHONE_OBJECT_TABLE IS

  BEGIN

    -- Return a table collection.
    RETURN SELF.telephone_entries;

  END;

  -- ------------------------------------------------------------

  -- A method to get a set member value.
  MEMBER FUNCTION get_set_member
  ( telephone_id         NUMBER   )
  RETURN TELEPHONE_OBJECT IS

    -- Define a local counter variable.
    c PLS_INTEGER := 1;

    -- Define a object structure.
    telephone TELEPHONE_OBJECT;
    
  BEGIN

    -- Use a for-loop to read through the set elements.
    FOR i IN 1..SELF.telephone_entries.LAST LOOP

      -- Check if set member is a match.
      IF SELF.telephone_entries(c).telephone_id = telephone_id THEN

        -- Assign object instance to object member.
        telephone := SELF.telephone_entries(c);

      END IF;

      -- Increment counter.
      c := c + 1;

    END LOOP;

    -- Return an instance of telephone.
    RETURN telephone;

  END;

  -- ------------------------------------------------------------

  -- Member procedure to delete a set of telephone objects.
  MEMBER PROCEDURE delete_telephone_object_set
  ( contact_id           NUMBER ) IS

    -- Define a local TELEPHONE_OBJECT variable.
    telephone            TELEPHONE_OBJECT;

    -- Define and initialize the object collection.
    telephones TELEPHONE_OBJECT_TABLE := telephone_object_table(); 

    -- Define a cursor to get set elements.
    CURSOR get_telephone
    ( contact_id_in IN NUMBER) IS
      SELECT   telephone_id
      FROM     telephone
      WHERE    contact_id = contact_id_in;

  BEGIN

    -- Use a for-loop to read through the set elements.
    FOR i IN get_telephone(contact_id) LOOP

      -- Initialize an object instance.
      telephone := telephone_object(i.telephone_id);

      -- Access the method of the object type.
      telephone.delete_telephone_object;
 
    END LOOP;

  END;

  -- ------------------------------------------------------------

  -- Member procedure to insert a set of telephone objects.
  MEMBER PROCEDURE insert_telephone_object_set
  ( telephone_table      TELEPHONE_OBJECT_TABLE ) IS

  BEGIN

    -- Loop through the initialized set.
    FOR i IN 1..SELF.TELEPHONE_ENTRIES.LAST LOOP

      -- Call the TELEPHONE_OBJECT method.
      SELF.telephone_entries(i).insert_telephone_object;

    END LOOP;

  END;

  -- ------------------------------------------------------------

  -- Member procedure to update a set of telephone objects.
  MEMBER PROCEDURE update_telephone_object_set
  ( telephone_entries    TELEPHONE_OBJECT_TABLE ) IS

  BEGIN

    -- Loop through the initialized set.
    FOR i IN 1..SELF.TELEPHONE_ENTRIES.LAST LOOP

      -- Call the TELEPHONE_OBJECT method.
      SELF.telephone_entries(i).update_telephone_object;

    END LOOP;

  END;

  -- ------------------------------------------------------------

  -- Member procedure to print the contents of the set.
  MEMBER PROCEDURE print_telephone_object_set IS

  BEGIN

    -- Print a double line break.
    DBMS_OUTPUT.PUT_LINE(
      '========================================');

    -- Use a for-loop to traverse the cursor.
    FOR i IN 1..SELF.telephone_entries.LAST LOOP

      -- Print a line break in-between return sets.
      IF i <> 1 THEN
        DBMS_OUTPUT.PUT_LINE(
          '----------------------------------------');
      END IF;

      -- Call the telephone_object print procedure.
      SELF.telephone_entries(i).print_telephone_object;        

    END LOOP;

    -- Print a double line break.
    DBMS_OUTPUT.PUT_LINE(
      '========================================');

  END;

  -- ------------------------------------------------------------

  -- Member procedure to print the phone numbers of the set.
  MEMBER PROCEDURE print_telephone_number_set IS

  BEGIN

    -- Print a double line break.
    DBMS_OUTPUT.PUT_LINE(
      '========================================');

    -- Use a for-loop to traverse the cursor.
    FOR i IN 1..SELF.telephone_entries.LAST LOOP

      -- Print a line break in-between return sets.
      IF i <> 1 THEN
        DBMS_OUTPUT.PUT_LINE(
          '----------------------------------------');
      END IF;

      -- Call the telephone_object print procedure.
      SELF.telephone_entries(i).print_telephone_number;        

    END LOOP;

    -- Print a double line break.
    DBMS_OUTPUT.PUT_LINE(
      '========================================');

  END;

  -- ------------------------------------------------------------

  -- Define a comparison function for a collection set object.
  ORDER MEMBER FUNCTION equals
  ( class_in TELEPHONE_OBJECT_SET )
  RETURN NUMBER IS

    -- Define return variable.
    retval      NUMBER := 1;

    -- Define a default return values.
    false_value NUMBER := 0;
    true_value  NUMBER := 1;

    -- Define local instances of telephone objects.
    telephones_1 TELEPHONE_OBJECT_TABLE := telephone_object_table(); 
    telephones_2 TELEPHONE_OBJECT_TABLE := telephone_object_table(); 

  BEGIN

    -- Assign set members to a table of objects.
    telephones_1 := SELF.get_collection;

    -- Assign set members to a table of objects.
    telephones_2 := class_in.get_collection;

    -- Loop through the initialized set.
    FOR i IN 1..SELF.TELEPHONE_ENTRIES.LAST LOOP

      -- Call the TELEPHONE_OBJECT method.
      IF telephones_1(i).equals(telephones_2(i)) = 1 THEN

        -- Assign a false value and exit loop.
        retval := false_value;

        -- Exit loop because one unequal element fails.
        EXIT;

      END IF;

    END LOOP;

    -- Return a default true numeric value.
    RETURN retval;

  END;

END;
/

SHOW ERRORS

SPOOL OFF
