/*
 * create_job1.sql
 * Chapter 7, Oracle Expert PL/SQL
 * by Ron Hardman and Michael McLaughlin
 *
 * This script demonstrates DBMS_JOB submission.
 */

CLEAR BREAKS
CLEAR COLUMNS
CLEAR COMPUTES

TTITLE OFF

SPOOL create_tables.log

SET ECHO ON
SET FEEDBACK ON
SET NULL '<Null>'
SET PAGESIZE 999
SET PAUSE OFF
SET SERVEROUTPUT ON SIZE 1000000
SET TERM ON
SET TIME OFF
SET TIMING OFF
SET VERIFY OFF

-- Cleanup the environment.
DECLARE

  -- Define a cursor to identify tables.
  CURSOR tables
  ( table_name_in VARCHAR2 ) IS
    SELECT   NULL
    FROM     user_tables
    WHERE    table_name = table_name_in;

  -- Define a cursor to identify tables.
  CURSOR sequences
  ( sequence_name_in VARCHAR2 ) IS
    SELECT   NULL
    FROM     user_sequences
    WHERE    sequence_name = sequence_name_in;

BEGIN

  -- Check for existing tables and drop them.
  FOR i IN tables('RESULT') LOOP
  
    -- Drop any duplicate table before attempting to create it.
    EXECUTE IMMEDIATE 'DROP TABLE result';
    
  END LOOP;

  -- Check for existing sequences and drop them.
  FOR i IN tables('RESULT_S1') LOOP
  
    -- Drop any duplicate sequence before attempting to create it.
    EXECUTE IMMEDIATE 'DROP SEQUENCE result';
    
  END LOOP;

END;
/

-- Create a result table.
CREATE TABLE result
( result_id    NUMBER
, submit       DATE
, process      DATE
, result_value NUMBER );

-- Create a sequence for the result table.
CREATE SEQUENCE result_s1
INCREMENT BY    1
START WITH      1
ORDER;

-- Create a package that shows a clock difference.
CREATE OR REPLACE PACKAGE sample_job1 IS

  -- Define a function with default values.
  FUNCTION add_loop
  ( minimum     NUMBER := 1
  , maximum     NUMBER := 10000000 )
  RETURN NUMBER;

  -- Define a procedure to display results.
  PROCEDURE display_results
  ( submit_date DATE
  , minimum     NUMBER := NULL
  , maximum     NUMBER := NULL );

END sample_job1;
/

SHOW ERRORS

-- Create a package that shows a clock difference.
CREATE OR REPLACE PACKAGE BODY sample_job1 IS

  -- Define a function with default values.
  FUNCTION add_loop
  ( minimum NUMBER := 1
  , maximum NUMBER := 10000000 )
  RETURN NUMBER IS

    -- Declare local variable.
    retval NUMBER := 0;

  BEGIN

    -- Loop from the minimum to maximum.
    FOR i IN minimum..maximum LOOP

      -- Check if the loop index is equal to the maximum value.
      IF maximum = i THEN

        -- Assign the value.
        retval := i;

      END IF;

    END LOOP;

    -- Return value.
    RETURN retval;

  END add_loop;

  -- Define a procedure to display results.
  PROCEDURE display_results
  ( submit_date DATE
  , minimum     NUMBER := NULL
  , maximum     NUMBER := NULL ) IS

    -- Set current time.
    current_date DATE := SYSDATE;

    -- Calculate the value.
    calculated_value NUMBER := sample_job1.add_loop;

  BEGIN

    -- Manage return value.
    INSERT
    INTO     result
    VALUES
    ( result_s1.nextval
    , submit_date
    , SYSDATE
    , calculated_value );

  END display_results;

END sample_job1;
/

SHOW ERRORS

DECLARE

  -- Record the job number.
  job_out PLS_INTEGER;

BEGIN

  -- Submit the job for immediate execution.
  DBMS_JOB.SUBMIT( JOB => job_out
                 , WHAT => 'sample_job1.display_results(SYSDATE);' );

  -- DBMS_JOB.SUBMIT requires a COMMIT to signal complete submission.
  COMMIT;

  -- Print the job number.
  DBMS_OUTPUT.PUT_LINE('The job is: ['||job_out||']');

END;
/

COL date1 FORMAT A20
COL date2 FORMAT A20

-- Query result table.
SELECT   result_id
,        TO_CHAR(submit,'DD-MON-YY HH24:MI:SSSS') date1
,        TO_CHAR(process,'DD-MON-YY HH24:MI:SSSS') date2
,        result_value
FROM     result;

-- Execute a null loop to wait for the job to complete.
BEGIN

  FOR i IN 1..100000000 LOOP
    NULL;
  END LOOP;

END;
/

-- Query result table.
SELECT   result_id
,        TO_CHAR(submit,'DD-MON-YY HH24:MI:SSSS') date1
,        TO_CHAR(process,'DD-MON-YY HH24:MI:SSSS') date2
,        result_value
FROM     result;
