/*
 * create_thread_of_control.sql
 * Chapter 7, Oracle Expert PL/SQL
 * by Ron Hardman and Michael McLaughlin
 *
 * This script builds a package to act as a thread-of-control
 * for parallel or concurrent PL/SQL programs.
 */

SET ECHO ON
SET SERVEROUTPUT ON SIZE 1000000

-- Call script to build local pipes.
@create_thread_pipes.sql

SPOOL audit.txt

BEGIN

  FOR i IN (SELECT   type_name
            FROM     user_types
            WHERE    type_name = 'JOBS_OBJECT' ) LOOP
            
    -- Use NDS to drop the dependent object type.
    EXECUTE IMMEDIATE 'DROP TYPE jobs_object';
    
  END LOOP;
  
  FOR i IN (SELECT   type_name
            FROM     user_types
            WHERE    type_name = 'JOBS_TABLE' ) LOOP
            
    -- Use NDS to drop the dependent object type.
    EXECUTE IMMEDIATE 'DROP TYPE jobs_table';
    
  END LOOP;
  
  FOR i IN (SELECT   type_name
            FROM     user_types
            WHERE    type_name = 'ATTRIBUTES_TABLE' ) LOOP
            
    -- Use NDS to drop the dependent object type.
    EXECUTE IMMEDIATE 'DROP TYPE attributes_table';
    
  END LOOP;

  FOR i IN (SELECT   type_name
            FROM     user_types
            WHERE    type_name = 'ATTRIBUTES_OBJECT' ) LOOP
            
    -- Use NDS to drop the dependent object type.
    EXECUTE IMMEDIATE 'DROP TYPE attributes_object';
    
  END LOOP;

  FOR i IN (SELECT   type_name
            FROM     user_types
            WHERE    type_name = 'ATTRIBUTE_TABLE' ) LOOP
            
    -- Use NDS to drop the dependent object type.
    EXECUTE IMMEDIATE 'DROP TYPE attribute_table';
    
  END LOOP;  
  
END;
/

CREATE OR REPLACE TYPE attribute_object AS OBJECT
( attribute_type     VARCHAR2(12)
, attribute_number   VARCHAR2(38)
, attribute_varchar2 VARCHAR2(100)
, CONSTRUCTOR FUNCTION attribute_object
  ( attribute_type   VARCHAR2
  , attribute_value  VARCHAR2 ) RETURN SELF AS RESULT
, MEMBER FUNCTION get_type RETURN VARCHAR2
, MEMBER FUNCTION get_number RETURN NUMBER
, MEMBER FUNCTION get_varchar2 RETURN VARCHAR2
, MEMBER FUNCTION to_varchar2 RETURN VARCHAR2
, ORDER MEMBER FUNCTION equals
  ( class_in         ATTRIBUTE_OBJECT ) RETURN NUMBER )
INSTANTIABLE FINAL;
/

show errors

CREATE OR REPLACE TYPE BODY attribute_object AS

  -- Constructor to build a null collection.
  CONSTRUCTOR FUNCTION attribute_object
  ( attribute_type  VARCHAR2
  , attribute_value VARCHAR2 ) RETURN SELF AS RESULT IS
 
  BEGIN

    -- Assign the type to the instance attribute.
    SELF.attribute_type := attribute_type;

    -- Check the type and build the variable type.    
    IF SELF.attribute_type = 'NUMBER' THEN
    
      SELF.attribute_number := attribute_value;
      SELF.attribute_varchar2 := NULL;
    
    ELSIF SELF.attribute_type = 'VARCHAR2' THEN
    
      SELF.attribute_number := 0;
      SELF.attribute_varchar2 := attribute_value;
        
    END IF;

    -- Return instance.
    RETURN;
  
  END;
  
  -- --------------------------------------------------------
     
  -- Member function to return attribute type.
  MEMBER FUNCTION get_type RETURN VARCHAR2 IS
  
  BEGIN
  
    -- Return the size of the collection.
    RETURN SELF.attribute_type;
  
  END;

  -- --------------------------------------------------------
     
  -- Member function to return attribute NUMBER.
  MEMBER FUNCTION get_number RETURN NUMBER IS
  
  BEGIN
  
    -- Return the element of the collection.
    RETURN SELF.attribute_number;
  
  END;

  -- --------------------------------------------------------
     
  -- Member function to return attribute VARCHAR2.
  MEMBER FUNCTION get_varchar2 RETURN VARCHAR2 IS
  
  BEGIN
  
    -- Return the element of the collection.
    RETURN SELF.attribute_varchar2;
  
  END;

  -- --------------------------------------------------------
     
  -- Member function to return attribute VARCHAR2.
  MEMBER FUNCTION to_varchar2 RETURN VARCHAR2 IS
  
  BEGIN
  
    IF SELF.attribute_number IS NULL THEN
    
      -- Return the element of the collection.
      RETURN SELF.attribute_varchar2;
      
    ELSE
    
      -- Return the element of the collection.
      RETURN SELF.attribute_number;
      
    END IF;
  
  END;

  -- ------------------------------------------------------------

  -- A method to check for equality of object instances.
  ORDER MEMBER FUNCTION equals
  ( class_in ATTRIBUTE_OBJECT )
  RETURN NUMBER IS

    -- Define a default return values.
    false_value NUMBER := 0;
    true_value  NUMBER := 1;

  BEGIN

    -- Compare all elements in the object.
    IF SELF.attribute_type = class_in.attribute_type AND
      SELF.attribute_number = class_in.attribute_number AND
      SELF.attribute_varchar2 = class_in.attribute_varchar2 THEN

      -- Return True they are equal.
      RETURN true_value;

    ELSE 

      -- Return False they are unequal.
      RETURN false_value;

    END IF;
      
  END;


END;
/

show errors

CREATE OR REPLACE TYPE attribute_table
  AS TABLE OF ATTRIBUTE_OBJECT;
/

show errors

-- Create object type definition or specification.
CREATE OR REPLACE TYPE attributes_object AS OBJECT
( attribute_list    ATTRIBUTE_TABLE
, CONSTRUCTOR FUNCTION attributes_object RETURN SELF AS RESULT
, CONSTRUCTOR FUNCTION attributes_object
  ( attribute_in    ATTRIBUTE_OBJECT ) RETURN SELF AS RESULT
, CONSTRUCTOR FUNCTION attributes_object
  ( attribute_list  ATTRIBUTE_TABLE ) RETURN SELF AS RESULT
, MEMBER PROCEDURE add_item
  ( attribute_in    ATTRIBUTE_OBJECT )
, MEMBER FUNCTION count RETURN NUMBER
, MEMBER PROCEDURE delete_item
  ( index_value     NUMBER )
, MEMBER PROCEDURE delete_item
  ( index_start     NUMBER
  , index_end       NUMBER )
, MEMBER FUNCTION limit RETURN NUMBER
, MEMBER FUNCTION get_attribute_list RETURN ATTRIBUTE_TABLE
, MEMBER FUNCTION to_varchar2 RETURN VARCHAR2
, ORDER MEMBER FUNCTION equals
  ( class_in        ATTRIBUTES_OBJECT ) RETURN NUMBER )
INSTANTIABLE FINAL;
/

show errors

-- Create object type implementation or body.
CREATE OR REPLACE TYPE BODY attributes_object AS

  -- Constructor to build a null collection.
  CONSTRUCTOR FUNCTION attributes_object RETURN SELF AS RESULT IS

  BEGIN

    -- Define collection variable.
    SELF.attribute_list  := attribute_table();

    -- Return instance.
    RETURN;
  
  END;
  
  -- Constructor to build a collection with an initial value.
  CONSTRUCTOR FUNCTION attributes_object
  ( attribute_in    ATTRIBUTE_OBJECT ) RETURN SELF AS RESULT IS

  BEGIN

    -- Initialize the instance variable.
    SELF.attribute_list := attribute_table(attribute_in);
  
    -- Return instance.
    RETURN;
  
  END;
  
  -- Constructor to build a collection with an initial collection.
  CONSTRUCTOR FUNCTION attributes_object
  ( attribute_list  ATTRIBUTE_TABLE ) RETURN SELF AS RESULT IS

  BEGIN

    -- Assign an existing collection to the instance collection.
    SELF.attribute_list := attribute_list;

    -- Return instance.
    RETURN;
  
  END;

  -- --------------------------------------------------------
     
  -- Member procedure to add to the internal collection.
  MEMBER PROCEDURE add_item
  ( attribute_in    ATTRIBUTE_OBJECT ) IS

    -- Define and initialize collection
    attribute_list ATTRIBUTE_TABLE := attribute_table();

  BEGIN

    -- Extend the instance collection space by one.
    SELF.attribute_list.EXTEND;

    -- Assign the new value.
    -- ----------------------------------------------------------
    --  The Collection API COUNT or LAST function will return
    --  the subscript for allocated space. If there where two
    --  items in the collection before the EXTEND statement
    --  above, the subscript returned by LAST would be three.
    -- ----------------------------------------------------------
    SELF.attribute_list(SELF.attribute_list.LAST) := attribute_in;

  END;

  -- --------------------------------------------------------
     
  -- Member function to return size of internal collection.
  MEMBER FUNCTION count RETURN NUMBER IS
  
  BEGIN
  
    -- Return the size of the collection.
    RETURN SELF.attribute_list.COUNT;
  
  END;

  -- --------------------------------------------------------
     
  -- Member function to return size of internal collection.
  MEMBER PROCEDURE delete_item
  ( index_value     NUMBER ) IS

    -- Declare a local counter variable.
    counter         NUMBER := 1;
    
    -- Define and initialize a local collection.
    attribute_list  ATTRIBUTE_TABLE := attribute_table();
  
  BEGIN
  
    -- Check for valid index value.      
    IF SELF.attribute_list.EXISTS(index_value) THEN

      -- Delete element.
      SELF.attribute_list.DELETE(index_value);

      -- Allocate space to the local collection.
      attribute_list.EXTEND(SELF.attribute_list.COUNT);

      -- Copy contents to a reindexed collection.
      FOR i IN SELF.attribute_list.FIRST..SELF.attribute_list.LAST LOOP
       
        -- Check if index value exists.
        IF SELF.attribute_list.EXISTS(i) THEN
      
          -- Assign a value to the reindexed collection.
          attribute_list(counter) := SELF.attribute_list(i); 
        
          -- Increment the local counter.
          counter := counter + 1;
    
        END IF;
    
      END LOOP;

    END IF;

    -- Assign the re-indexed collection to the instance collection.
    SELF.attribute_list := attribute_list;

  END;

  -- --------------------------------------------------------
     
  -- Member function to return size of internal collection.
  MEMBER PROCEDURE delete_item
  ( index_start     NUMBER
  , index_end       NUMBER ) IS
  
    -- Declare a local counter variable.
    counter         NUMBER := 1;
   
    -- Define and initialize a local collection.
    attribute_list  ATTRIBUTE_TABLE := attribute_table();
  
  BEGIN

    -- Check starting index is less than the ending index.
    IF index_start >= index_end THEN
    
      RAISE_APPLICATION_ERROR(-20000,
        'Starting index is larger than ending index');
      
    END IF;
  
    -- Check for a valid index.
    IF SELF.attribute_list.EXISTS(index_start) AND
       SELF.attribute_list.EXISTS(index_end) THEN

      -- Delete element.
      SELF.attribute_list.DELETE(index_start,index_end);

      -- Allocate space to the local collection.
      attribute_list.EXTEND(SELF.attribute_list.COUNT);    
    
      -- Copy contents to a reindexed collection.
      FOR i IN SELF.attribute_list.FIRST..SELF.attribute_list.LAST LOOP
    
        -- Check if index value exists.
        IF SELF.attribute_list.EXISTS(i) THEN
   
          -- Assign a value to the reindexed collection.
          attribute_list(counter) := SELF.attribute_list(i); 
        
          -- Increment the local counter.
          counter := counter + 1;
    
        END IF;

      END LOOP;

    END IF;
    
    -- Assign the re-indexed collection to the instance collection.
    SELF.attribute_list := attribute_list;
    
  END;

  -- --------------------------------------------------------
     
  -- Member function to return size of internal collection.
  MEMBER FUNCTION limit RETURN NUMBER IS
  
  BEGIN
  
    -- Return the size of the collection.
    RETURN SELF.attribute_list.COUNT;
  
  END;

  -- --------------------------------------------------------
     
  -- Member function to return the internal collection.
  MEMBER FUNCTION get_attribute_list
  RETURN ATTRIBUTE_TABLE IS
  
  BEGIN
  
    -- Return a table collection.
    RETURN SELF.attribute_list;
  
  END;

  -- --------------------------------------------------------
     
  -- Member function to return attribute VARCHAR2.
  MEMBER FUNCTION to_varchar2 RETURN VARCHAR2 IS
  
    -- List the attributes.
    list            VARCHAR2(4000);
  
  BEGIN

    -- Read and serialize values.
    FOR i IN 1..SELF.attribute_list.COUNT LOOP

      IF i = 1 THEN

        -- Assign values.
        list :=
          '['||i||'] ['||SELF.attribute_list(i).to_varchar2||']'||CHR(10);

      ELSIF i = SELF.attribute_list.COUNT THEN
      
        -- Assign values.
        list := list ||
          '['||i||'] ['||SELF.attribute_list(i).to_varchar2||']';
      
      ELSE
      
        -- Assign values.
        list := list ||
          '['||i||'] ['||SELF.attribute_list(i).to_varchar2||']'||CHR(10);
      
      END IF;
      
    END LOOP;

    -- Return the size of the collection.
    RETURN list;
  
  END;

  -- --------------------------------------------------------
     
  -- Order method to compare current instance with another.
  ORDER MEMBER FUNCTION equals
  ( class_in ATTRIBUTES_OBJECT ) RETURN NUMBER IS
  
    -- Define return variable.
    retval          NUMBER := 1;

    -- Define a default return values.
    false_value     NUMBER := 0;
    true_value      NUMBER := 1;

    -- Define local instances of telephone objects.
    attributes_1    ATTRIBUTE_TABLE := attribute_table(); 
    attributes_2    ATTRIBUTE_TABLE := attribute_table(); 

  BEGIN

    -- Assign set members to a table of objects.
    attributes_1 := SELF.get_attribute_list;

    -- Assign set members to a table of objects.
    attributes_2 := class_in.get_attribute_list;

    -- Loop through the initialized set.
    FOR i IN 1..attributes_1.LAST LOOP

      -- Call the JOBS_OBJECT method.
      IF attributes_1(i) = attributes_2(i) THEN

        -- Assign a false value and exit loop.
        retval := false_value;

        -- Exit loop because one unequal element fails.
        EXIT;

      END IF;

    END LOOP;

    -- Return a default true numeric value.
    RETURN retval;
    
  END;
  
END;
/

show errors

CREATE OR REPLACE TYPE attributes_table
  AS TABLE OF ATTRIBUTES_OBJECT;
/

show errors

-- Create a JOB_TABLE collection of numbers.
CREATE OR REPLACE TYPE job_table
  AS TABLE OF NUMBER;
/

-- Create object type definition or specification.
CREATE OR REPLACE TYPE jobs_object AS OBJECT
( job_list          JOB_TABLE
, CONSTRUCTOR FUNCTION jobs_object RETURN SELF AS RESULT
, CONSTRUCTOR FUNCTION jobs_object
  ( job_number      NUMBER ) RETURN SELF AS RESULT
, CONSTRUCTOR FUNCTION jobs_object
  ( job_list_in     JOB_TABLE ) RETURN SELF AS RESULT
, MEMBER PROCEDURE add_item
  ( job_number      NUMBER )
, MEMBER FUNCTION count RETURN NUMBER
, MEMBER PROCEDURE delete_item
  ( index_value     NUMBER )
, MEMBER PROCEDURE delete_item
  ( index_start     NUMBER
  , index_end       NUMBER )
, MEMBER FUNCTION limit RETURN NUMBER
, MEMBER FUNCTION get_job_list RETURN JOB_TABLE
, ORDER MEMBER FUNCTION equals
  ( class_in        JOBS_OBJECT ) RETURN NUMBER )
INSTANTIABLE FINAL;
/

show errors

-- Create object type implementation or body.
CREATE OR REPLACE TYPE BODY jobs_object AS

  -- Constructor to build a null collection.
  CONSTRUCTOR FUNCTION jobs_object RETURN SELF AS RESULT IS
 
  BEGIN

    -- Initialize the instance variable.
    SELF.job_list := job_table();

    -- Return instance.
    RETURN;
  
  END;
  
  -- Constructor to build a collection with an initial value.
  CONSTRUCTOR FUNCTION jobs_object
  ( job_number      NUMBER ) RETURN SELF AS RESULT IS

  BEGIN

    -- Initialize the instance variable.
    SELF.job_list := job_table(job_number);
  
    -- Return instance.
    RETURN;
  
  END;
  
  -- Constructor to build a collection with an initial collection.
  CONSTRUCTOR FUNCTION jobs_object
  ( job_list_in     JOB_TABLE ) RETURN SELF AS RESULT IS

  BEGIN

    -- Assign an existing collection to the instance collection.
    SELF.job_list := job_list_in;

    -- Return instance.
    RETURN;
  
  END;

  -- --------------------------------------------------------
     
  -- Member procedure to add to the internal collection.
  MEMBER PROCEDURE add_item
  ( job_number      NUMBER ) IS

  BEGIN

   -- SELF.job_list := job_table();

    -- Extend the instance collection space by one.
    SELF.job_list.EXTEND;

    -- Assign the new value.
    -- ----------------------------------------------------------
    --  The Collection API COUNT or LAST function will return
    --  the subscript for allocated space. If there where two
    --  items in the collection before the EXTEND statement
    --  above, the subscript returned by LAST would be three.
    -- ----------------------------------------------------------
    SELF.job_list(SELF.job_list.LAST) := job_number;

  END;

  -- --------------------------------------------------------
     
  -- Member function to return size of internal collection.
  MEMBER FUNCTION count RETURN NUMBER IS
  
  BEGIN
  
    -- Return the size of the collection.
    RETURN SELF.job_list.COUNT;
  
  END;

  -- --------------------------------------------------------
     
  -- Member function to return size of internal collection.
  MEMBER PROCEDURE delete_item
  ( index_value     NUMBER ) IS

    -- Declare a local counter variable.
    counter         NUMBER := 1;
    
    -- Define and initialize a local collection.
    job_list        JOB_TABLE := job_table();
  
  BEGIN
  
    -- Check for valid index value.      
    IF SELF.job_list.EXISTS(index_value) THEN

      -- Delete element.
      SELF.job_list.DELETE(index_value);

      -- Allocate space to the local collection.
      job_list.EXTEND(SELF.job_list.COUNT);

      -- Copy contents to a reindexed collection.
      FOR i IN SELF.job_list.FIRST..SELF.job_list.LAST LOOP
       
        -- Check if index value exists.
        IF SELF.job_list.EXISTS(i) THEN
      
          -- Assign a value to the reindexed collection.
          job_list(counter) := SELF.job_list(i); 
        
          -- Increment the local counter.
          counter := counter + 1;
    
        END IF;
    
      END LOOP;

    END IF;

    -- Assign the re-indexed collection to the instance collection.
    SELF.job_list := job_list;

  END;

  -- --------------------------------------------------------
     
  -- Member function to return size of internal collection.
  MEMBER PROCEDURE delete_item
  ( index_start     NUMBER
  , index_end       NUMBER ) IS
  
    -- Declare a local counter variable.
    counter         NUMBER := 1;
   
    -- Define and initialize a local collection.
    job_list        JOB_TABLE := job_table();
  
  BEGIN

    -- Check starting index is less than the ending index.
    IF index_start >= index_end THEN
    
      RAISE_APPLICATION_ERROR(-20000,
        'Starting index is larger than ending index');
      
    END IF;
  
    -- Check for a valid index.
    IF SELF.job_list.EXISTS(index_start) AND
       SELF.job_list.EXISTS(index_end) THEN

      -- Delete element.
      SELF.job_list.DELETE(index_start,index_end);

      -- Allocate space to the local collection.
      job_list.EXTEND(SELF.job_list.COUNT);    
    
      -- Copy contents to a reindexed collection.
      FOR i IN SELF.job_list.FIRST..SELF.job_list.LAST LOOP
    
        -- Check if index value exists.
        IF SELF.job_list.EXISTS(i) THEN
   
          -- Assign a value to the reindexed collection.
          job_list(counter) := SELF.job_list(i); 
        
          -- Increment the local counter.
          counter := counter + 1;
    
        END IF;

      END LOOP;

    END IF;
    
    -- Assign the re-indexed collection to the instance collection.
    SELF.job_list := job_list;
    
  END;

  -- --------------------------------------------------------
     
  -- Member function to return size of internal collection.
  MEMBER FUNCTION limit RETURN NUMBER IS
  
  BEGIN
  
    -- Return the size of the collection.
    RETURN SELF.job_list.COUNT;
  
  END;

  -- --------------------------------------------------------
     
  -- Member function to return the internal collection.
  MEMBER FUNCTION get_job_list
  RETURN JOB_TABLE IS
  
  BEGIN
  
    -- Return a table collection.
    RETURN SELF.job_list;
  
  END;

  -- --------------------------------------------------------
     
  -- Order method to compare current instance with another.
  ORDER MEMBER FUNCTION equals
  ( class_in JOBS_OBJECT ) RETURN NUMBER IS
  
    -- Define return variable.
    retval          NUMBER := 1;

    -- Define a default return values.
    false_value     NUMBER := 0;
    true_value      NUMBER := 1;

    -- Define local instances of telephone objects.
    jobs_1          JOB_TABLE := job_table(); 
    jobs_2          JOB_TABLE := job_table(); 

  BEGIN

    -- Assign set members to a table of objects.
    jobs_1 := SELF.get_job_list;

    -- Assign set members to a table of objects.
    jobs_2 := class_in.get_job_list;

    -- Loop through the initialized set.
    FOR i IN 1..jobs_1.LAST LOOP

      -- Call the JOBS_OBJECT method.
      IF jobs_1(i) = jobs_2(i) THEN

        -- Assign a false value and exit loop.
        retval := false_value;

        -- Exit loop because one unequal element fails.
        EXIT;

      END IF;

    END LOOP;

    -- Return a default true numeric value.
    RETURN retval;
    
  END;
  
END;
/

show errors

-- Create package specification.
CREATE OR REPLACE PACKAGE thread_of_control IS

  -- Define a job message type.
  TYPE job_message IS TABLE OF VARCHAR2(100); 
    
  -- Define standard exceptions.
  complete_failed        EXCEPTION;
  lock_failed            EXCEPTION;
  send_failed            EXCEPTION;
  submit_failed          EXCEPTION;
  receive_failed         EXCEPTION;
  rewrite_failed         EXCEPTION;
  rewrite_lock_failed    EXCEPTION;

  -- Define the PRAGMA for user-defined exceptions.
  PRAGMA EXCEPTION_INIT(complete_failed,-200001);
  PRAGMA EXCEPTION_INIT(lock_failed,-200002);
  PRAGMA EXCEPTION_INIT(send_failed,-200003);
  PRAGMA EXCEPTION_INIT(submit_failed,-200004);
  PRAGMA EXCEPTION_INIT(receive_failed,-200005);
  PRAGMA EXCEPTION_INIT(rewrite_failed,-200006);
  PRAGMA EXCEPTION_INIT(rewrite_lock_failed,-200007);

  -- Define manage_threads function specification.
  FUNCTION manage_threads
    ( user_name     IN VARCHAR2
    , program_name  IN VARCHAR2
    , attributes    IN ATTRIBUTES_TABLE
    , run_threads   IN NUMBER := 1 )
    RETURN BOOLEAN;

END thread_of_control;
/

-- Show any errors while building specification.
show errors

-- Create package body.
CREATE OR REPLACE PACKAGE BODY thread_of_control IS

  -- Define a package variable.
  locked       VARCHAR2(4)   := 'LOCK';

  -- Define a forward references to alphabetize local units.
  PROCEDURE register_threads
    ( user_name     IN     VARCHAR2
    , program_name  IN     VARCHAR2
    , attributes    IN     ATTRIBUTES_TABLE
    , job_list      IN OUT JOBS_OBJECT
    , run_threads   IN     NUMBER := 1 );

  FUNCTION send_message
    ( user_name     VARCHAR2
    , message_in    VARCHAR2
    , pipe_name     VARCHAR2 ) RETURN INTEGER;
    
  FUNCTION receive_message
    ( user_name     VARCHAR2
    , pipe_name     VARCHAR2 ) RETURN VARCHAR2;

  PROCEDURE wait;

  -- --------------------------------------------------------
     
  -- Define complete_threads function.
  PROCEDURE complete_threads
    ( user_name      IN     VARCHAR2
    , program_name   IN     VARCHAR2
    , job_list       IN OUT JOBS_OBJECT
    , global_commit  IN OUT BOOLEAN ) IS

    -- Define local variables.
    message         VARCHAR2(4000);
    
    -- Define structures to hold contents of named pipe.
    job_message_list JOB_MESSAGE := job_message();
    
  BEGIN

    -- Set the lock.
    IF (thread_of_control.receive_message
          ( user_name
          ,'THREAD_REGISTER_LOCK' ) = locked ) THEN

       -- Read complete list.
       FOR i IN 1..job_list.LIMIT LOOP

         -- Assign the message.
         message := thread_of_control.receive_message
                      ( user_name
                      ,'THREAD_REGISTER');

         -- Check for status of threads.                 
         IF (program_name = SUBSTR(message,1,INSTR(message,'[',1,1) - 1)) THEN

           -- Extend collection.    
           job_message_list.EXTEND;

           -- Assign the message to the list.
           job_message_list(i) := message;

           -- Check if all threads are complete.
           IF job_message_list.COUNT = job_list.LIMIT THEN

             -- Update thread object type as complete.         
             global_commit := TRUE;
       
           END IF;

         -- Check for a partial completion of threads.
         ELSIF   job_message_list.COUNT < job_list.LIMIT
         AND     ( message = 'The message pipe is empty.'
         OR        message = 'The message is too large for variable.'
         OR        message = 'An interrupt occurred, contact the DBA.' ) THEN

           -- Loop thru what has been read and write it back to the pipe.
           FOR j IN 1..job_message_list.COUNT LOOP
         
             -- Check re-registering of threads in the pipe.
             IF (thread_of_control.send_message
                   ( user_name
                   , job_message_list(j)
                   ,'THREAD_REGISTER' ) != 0) THEN
 
               -- Raise exception.
               RAISE rewrite_failed;               

             END IF;
             
           END LOOP;

           -- Set the lock.
           IF (thread_of_control.send_message
                 ( user_name
                 , locked
                 ,'THREAD_REGISTER_LOCK' ) != 0) THEN

              -- Raise exception.
              RAISE rewrite_lock_failed;
                                      
           END IF;
             
         END IF;
                 
       END LOOP;
    
       -- Verify that all threads have completed.
       IF job_list.LIMIT != job_message_list.COUNT THEN

        -- Raise exception.
        RAISE rewrite_failed;
                 
       END IF;
           
    END IF;
        
  EXCEPTION
  
    WHEN complete_failed THEN
      DBMS_OUTPUT.PUT_LINE('Complete failed: ['||SQLERRM||']');
      RETURN;
   
    WHEN rewrite_failed THEN
      DBMS_OUTPUT.PUT_LINE('Rewrite ['||message||
                           '] failed: ['||SQLERRM||']');
      RETURN;

    WHEN rewrite_lock_failed THEN
      DBMS_OUTPUT.PUT_LINE('Rewrite lock ['||message||
                           '_LOCK ] failed: ['||SQLERRM||']');
      RETURN;
   
  END complete_threads;

  -- --------------------------------------------------------
     
  -- Define manage_threads function.
  FUNCTION manage_threads
    ( user_name     IN VARCHAR2
    , program_name  IN VARCHAR2
    , attributes    IN ATTRIBUTES_TABLE
    , run_threads   IN NUMBER := 1)
    RETURN BOOLEAN IS
    
  -- Define a process complete Boolean variable.
  process_complete   BOOLEAN := FALSE;
  retval             BOOLEAN := FALSE;

  -- Define collection job set.
  job_set           JOB_TABLE := job_table();

  -- Define collection variable.
  job_list          JOBS_OBJECT := jobs_object();
 
  BEGIN

    -- If one thread fails all should fail.  
    thread_of_control.register_threads( user_name
                                      , program_name
                                      , attributes
                                      , job_list
                                      , run_threads );

    IF job_list.LIMIT > 0 THEN
  
      -- Assign the object's internal collection to a local variable.
      job_set := job_list.get_job_list;

    END IF;
    
                                      
    FOR i IN 1..attributes.COUNT LOOP
    
      IF job_list.LIMIT > 0 THEN
    
        -- Commit job submissions.
        COMMIT;
                                           
        -- Loop until all threads are complete.      
        WHILE NOT process_complete LOOP

          -- Wait before checking or rechecking.
          wait;

          -- Check if all threads are complete. 
          complete_threads ( user_name
                           , program_name
                           , job_list
                           , process_complete );

          -- Check if global commit has happened.
          IF process_complete THEN

            -- Exit the while loop.        
            EXIT;
          
          END IF;
                                     
        END LOOP;

      END IF;

      -- Return a Boolean completion of true.
      retval := TRUE;

    END LOOP;

    -- Return result.
    RETURN retval;

           
  END manage_threads;

  -- --------------------------------------------------------
     
  -- Implement package function defined in specification.
  FUNCTION receive_message
    ( user_name      VARCHAR2
    , pipe_name      VARCHAR2 )
    RETURN VARCHAR2 IS

    -- Define variable for target mailbox.
    message         VARCHAR2(4000 CHAR) :=  NULL;
    target_pipe     VARCHAR2(100 CHAR);
    timeout         INTEGER := 0;
    return_code     INTEGER;

  BEGIN

    -- Purge local pipe content.
    DBMS_PIPE.RESET_BUFFER;

    -- Declare the target for a pipe.
    target_pipe := UPPER(user_name) || '$'
                || UPPER(pipe_name);

    -- Use the procedure to put a message in the local buffer. 
    return_code := DBMS_PIPE.receive_message(target_pipe,timeout);

    -- Evaluate and process return code.
    CASE return_code
      WHEN 0 THEN 

        -- Read the message into a variable.
        DBMS_PIPE.UNPACK_MESSAGE(message);

      WHEN 1 THEN

      -- Assign message.
      message := 'The message pipe is empty.';

      WHEN 2 THEN 

      -- Assign message.
      message := 'The message is too large for variable.';

      WHEN 3 THEN 

      -- Assign message.
      message := 'An interrupt occurred, contact the DBA.';

    END CASE;

    -- Return the message.
    RETURN message;

  END receive_message;

  -- --------------------------------------------------------
     
  -- Define register_threads function.
  PROCEDURE register_threads
    ( user_name     IN     VARCHAR2
    , program_name  IN     VARCHAR2
    , attributes    IN     ATTRIBUTES_TABLE
    , job_list      IN OUT JOBS_OBJECT
    , run_threads   IN     NUMBER := 1 ) IS

    -- Define argument list.
    arguments       VARCHAR2(4000);

    -- Define a variable to hold the job number.
    job_number      NUMBER;

    -- Define a variable to hold dynamic what execution.
    job_what        VARCHAR2(100);
  
    FUNCTION parse_arguments
    ( attributes    IN     ATTRIBUTES_OBJECT )
    RETURN VARCHAR2 IS

      -- Define a local variable to manage the argument list.
      args           VARCHAR2(4000);
      
      -- Define a attribute object table.
      attribute_list ATTRIBUTE_TABLE := attributes.get_attribute_list;
    
    BEGIN
    
      -- Loop through the attribute list.
      FOR i IN 1..attribute_list.LAST LOOP

        -- Check if the attribute is a number.  
        IF attribute_list(i).get_type = 'NUMBER' THEN

          -- Check if the first argument.
          IF i = 1 THEN
        
            -- Write to the argument list.
            args := args||attribute_list(i).get_number;
            
          ELSE
          
            -- Write to the argument list.
            args := args||','||attribute_list(i).get_number;
          
          END IF;
          
        ELSIF attribute_list(i).get_type = 'STRING' THEN

          -- Check if the first argument.
          IF i = 1 THEN
                 
            -- Write to the argument list.
            args := args||''''||attribute_list(i).get_varchar2||'''';
            
          ELSE
          
            -- Write to the argument list.
            args := args||','''||attribute_list(i).get_varchar2||'''';
            
          END IF;
  
        END IF;
  
      END LOOP;

      -- Return argument list as a VARCHAR2.    
      RETURN args;
    
    END;
  
  BEGIN

    -- Run threads with different formal parameter signatures.
    FOR i IN 1..attributes.COUNT LOOP
    
      -- Parse the argument list.
      arguments := parse_arguments(attributes(i));

      -- Run threads with identical formal parameter signatures.
      FOR j IN 1..run_threads LOOP
    
        -- Define thread name.    
        job_what := program_name||'['||i||'] ['||j||']';
    
        -- Submit the job for immediate execution.
        DBMS_JOB.SUBMIT( JOB => job_number
                       , WHAT => program_name||'('||arguments||');');

        -- Validate a job number is returned from DBMS_JOB.SUBMIT.
        IF job_number IS NOT NULL THEN

          -- Add the thread to the job list.
          job_list.add_item(job_number);

        ELSE
      
          -- Raise exception.
          RAISE submit_failed;
      
        END IF;

        -- Register thread.
        IF (thread_of_control.send_message( user_name
                                          , job_what
                                          ,'THREAD_REGISTER' ) != 0) THEN
          -- Raise exception.
          RAISE send_failed;
      
        END IF;

      END LOOP;

      -- Set the lock.
      IF (thread_of_control.send_message( user_name
                                        , locked
                                        ,'THREAD_REGISTER_LOCK' ) != 0) THEN

        -- Raise exception.
        RAISE lock_failed;
                                      
      END IF;
    
    END LOOP;
    
  EXCEPTION
  
    WHEN send_failed OR lock_failed THEN
      DBMS_OUTPUT.PUT_LINE('Write failed: ['||SQLERRM||']');
      RETURN;

  END register_threads;

  -- --------------------------------------------------------
     
  -- Implement local function.
  FUNCTION send_message
    ( user_name      VARCHAR2
    , message_in     VARCHAR2
    , pipe_name      VARCHAR2 )
    RETURN INTEGER IS

    -- Define variable for target.
    target_pipe     VARCHAR2(100 CHAR);

  BEGIN

    -- Purge local pipe content.
    DBMS_PIPE.RESET_BUFFER;

    -- Declare the target for a message.
    target_pipe := UPPER(user_name) || '$'
                || UPPER(pipe_name);

    -- Use the procedure to put a message in the local buffer. 
    DBMS_PIPE.PACK_MESSAGE(message_in);

    -- Send message, success is a zero return value.
    IF (DBMS_PIPE.send_message(target_pipe) = 0) THEN

      -- Message sent, so return 0.
      RETURN 0;

    ELSE

      -- Message not sent, so return 1.
      RETURN 1;

    END IF; 

  END send_message;

  -- --------------------------------------------------------
     
  -- Define wait function specification.
  PROCEDURE wait IS
    
  BEGIN
  
    -- Pause program in a loop.
    FOR i IN 1..10000000 LOOP
        
      NULL;
          
    END LOOP; 
    
  END wait;

  -- --------------------------------------------------------
     
END thread_of_control;
/

COL object_name FORMAT A30
COL status      FORMAT A7

SELECT   object_name
,        status
FROM     user_objects
WHERE    object_name IN ('ATTRIBUTE_OBJECT'
                        ,'ATTRIBUTE_TABLE'
                        ,'ATTRIBUTES_OBJECT'
                        ,'ATTRIBUTES_TABLE'
                        ,'JOBS_OBJECT'
                        ,'JOBS_TABLE'
                        ,'THREAD_OF_CONTROL');

-- Show any errors while building specification.
show errors

SPOOL OFF