/*
 * create_threading_objects.sql
 * Chapter 7, Oracle Expert PL/SQL
 * by Ron Hardman and Michael McLaughlin
 *
 * This script deletes pipes if it exists in the context of the current
 * session, then recreates them.
 */

SET ECHO ON
SET SERVEROUTPUT ON SIZE 1000000

-- Cleanup the environment.
BEGIN

  FOR i IN (SELECT   table_name
            FROM     user_tables
            WHERE    table_name = 'THREAD' ) LOOP
            
    -- Use NDS to drop the dependent object type.
    EXECUTE IMMEDIATE 'DROP TABLE thread';
    
  END LOOP;
  
END;
/

-- Build a target table.
CREATE TABLE thread
( thread_id         NUMBER
, value_start       NUMBER
, value_end         NUMBER
, date_created      DATE
, date_updated      DATE );

DECLARE

  -- Define a index-by tables (9i) or an associative arrays (10g).
  TYPE number_table IS TABLE OF NUMBER
    INDEX BY BINARY_INTEGER;
  TYPE date_table IS TABLE OF DATE
    INDEX BY BINARY_INTEGER;

  -- Define and initialize a static date.
  date_value      DATE := SYSDATE;

  -- Define three variables of NUMBER_TABLE.
  number1         NUMBER_TABLE;
  number2         NUMBER_TABLE;
  number3         NUMBER_TABLE;
  date1           DATE_TABLE;
  date2           DATE_TABLE;
    
BEGIN

  -- Initialize the index-by table.
  FOR i IN 1..1000000 LOOP
 
    number1(i) := i;
    number2(i) := 1;
    number3(i) := 100;
    date1(i) := date_value;
    date2(i) := date_value;
  
  END LOOP;

  -- Use a bulk insert.
  FORALL i IN 1..1000000
    INSERT
    INTO      thread
    VALUES (number1(i),number2(i),number3(i),date1(i),date2(i));

  -- Commit the inserts.  
  COMMIT;
  
END;
/

-- Define a stored procedure as a target for parallel processing.
CREATE OR REPLACE PROCEDURE threaded_program
( value_start_in     NUMBER
, value_end_in       NUMBER
, value_new_in       NUMBER ) IS

BEGIN

  -- Update rows.
  UPDATE   thread
  SET      value_end = value_new_in
  ,        date_updated = SYSDATE
  WHERE    thread_id BETWEEN value_start_in AND value_end_in;
  
END;
/

show errors

COL object_name FORMAT A30
COL status      FORMAT A7

SELECT   object_name
,        status
FROM     user_objects
WHERE    object_name IN ('ATTRIBUTE_OBJECT'
                        ,'ATTRIBUTE_TABLE'
                        ,'ATTRIBUTES_OBJECT'
                        ,'ATTRIBUTES_TABLE'
                        ,'JOBS_OBJECT'
                        ,'JOBS_TABLE'
                        ,'THREAD_OF_CONTROL'
                        ,'THREADED_PROGRAM');

DECLARE

  -- Define and initialize thread of execution variables.
  user_name         VARCHAR2(30) := USER;
  program_name      VARCHAR2(30) := 'THREADED_PROGRAM';
  attribute_count   NUMBER := 3;

  -- Define an ATTRIBUTES_OBJECT variable.
  attribute       ATTRIBUTES_OBJECT;
  
  -- Define the number of threads.
  threads           NUMBER := 2;

  -- Define collection types.
  attribute_list    ATTRIBUTE_TABLE := attribute_table();
  attributes_list   ATTRIBUTES_TABLE := attributes_table();
  
BEGIN

  -- Allocate space to the collections.
  attribute_list.EXTEND(attribute_count);
  attributes_list.EXTEND(threads);
 
  -- Use a for-loop to populate two threads of execution.
  FOR i IN 1..2 LOOP
  
    IF i = 1 THEN
  
      -- Initialize argument components.
      attribute_list(1) := attribute_object('NUMBER',1);
      attribute_list(2) := attribute_object('NUMBER',100);
      attribute_list(3) := attribute_object('NUMBER',333);

    ELSIF i = 2 THEN

      -- Initialize argument components.
      attribute_list(1) := attribute_object('NUMBER',201);
      attribute_list(2) := attribute_object('NUMBER',300);
      attribute_list(3) := attribute_object('NUMBER',444);

    END IF;

    -- Construct an instance of attributes.
    attribute := attributes_object(attribute_list);

    -- Assign element to collection.
    attributes_list(i) := attribute;

    dbms_output.put_line('----------------------------');
    dbms_output.put_line(attributes_list(i).to_varchar2);
    dbms_output.put_line('----------------------------');

  END LOOP;

  -- Submit the threads.
  IF THREAD_OF_CONTROL.MANAGE_THREADS( user_name
                                     , program_name
                                     , attributes_list ) THEN

    NULL;
                                     
  END IF;

END;
/  

-- Check all packages.
SELECT   object_name
,        status
FROM     user_objects
WHERE    object_name IN ('ATTRIBUTE_OBJECT'
                        ,'ATTRIBUTE_TABLE'
                        ,'ATTRIBUTES_OBJECT'
                        ,'ATTRIBUTES_TABLE'
                        ,'JOBS_OBJECT'
                        ,'JOBS_TABLE'
                        ,'THREAD_OF_CONTROL'
                        ,'THREADED_PROGRAM');