/*
 * create_wrapper.sql
 * Chapter 7, Oracle Expert PL/SQL
 * by Ron Hardman and Michael McLaughlin
 *
 * This script builds a package to send and receive
 * messages between programs.
 */

SET ECHO ON
SET SERVEROUTPUT ON SIZE 1000000

-- Call script to build local pipes.
@create_thread_pipes.sql

-- Create package specification.
CREATE OR REPLACE PACKAGE messenger IS

  -- Define function specification.
  FUNCTION send_message
    (user_name      VARCHAR2
    ,message_in     VARCHAR2
    ,pipe_name      VARCHAR2 )
    RETURN INTEGER;

  -- Define function specification.
  FUNCTION receive_message
    (user_name      VARCHAR2
    ,message_out    VARCHAR2
    ,pipe_name      VARCHAR2 )
    RETURN VARCHAR2;

END messenger;
/

-- Show any errors while building specification.
show errors

-- Create package body.
CREATE OR REPLACE PACKAGE BODY messenger IS

  -- Define local package function to return user name.
  FUNCTION get_user 
    RETURN VARCHAR2 IS

  BEGIN

    -- Use a cursor for-loop to get user name.
    FOR i IN (SELECT user FROM dual) LOOP

      -- Return the user.
      return i.user;

    END LOOP;
 
  END get_user;
  
  -- Implement package function defined in specification.
  FUNCTION send_message
    (user_name      VARCHAR2
    ,message_in     VARCHAR2
    ,pipe_name      VARCHAR2 )
    RETURN INTEGER IS

    -- Define variable for target.
    target_pipe     VARCHAR2(100 CHAR);

  BEGIN

    -- Purge local pipe content.
    DBMS_PIPE.RESET_BUFFER;

    -- Declare the target for a message.
    target_pipe := UPPER(user_name) || '$'
                || UPPER(pipe_name);

    -- Use the procedure to put a message in the local buffer. 
    DBMS_PIPE.PACK_MESSAGE(message_in);

    -- Send message, success is a zero return value.
    IF (DBMS_PIPE.send_message(target_pipe) = 0) THEN

      -- Message sent, so return 0.
      RETURN 0;

    ELSE

      -- Message not sent, so return 1.
      RETURN 1;

    END IF; 

  END send_message;
  
  -- Implement package function defined in specification.
  FUNCTION receive_message
    (user_name      VARCHAR2
    ,message_out    VARCHAR2
    ,pipe_name      VARCHAR2 )
    RETURN VARCHAR2 IS

    -- Define variable for target mailbox.
    message         VARCHAR2(4000 CHAR) :=  NULL;
    target_pipe     VARCHAR2(100 CHAR);
    timeout         INTEGER := 0;
    return_code     INTEGER;

  BEGIN

    -- Purge local pipe content.
    DBMS_PIPE.RESET_BUFFER;

    -- Declare the target for a pipe.
    target_pipe := UPPER(user_name) || '$'
                || UPPER(pipe_name);

    -- Use the procedure to put a message in the local buffer. 
    return_code := DBMS_PIPE.receive_message(target_pipe,timeout);

    -- Evaluate and process return code.
    CASE return_code
      WHEN 0 THEN 

        -- Read the message into a variable.
        DBMS_PIPE.UNPACK_MESSAGE(message);

      WHEN 1 THEN

      -- Assign message.
      message := 'The message pipe is empty.';

      WHEN 2 THEN 

      -- Assign message.
      message := 'The message is too large for variable.';

      WHEN 3 THEN 

      -- Assign message.
      message := 'An interrupt occurred, contact the DBA.';

    END CASE;

    -- Return the message.
    RETURN message;

  END receive_message;

END messenger;
/

-- Show any errors while building specification.
show errors
