/*
 * create_user.sql
 * Chapter 9, Expert PL/SQL Programming
 * by Ron Hardman and Mike McLaughlin
 *
 * This script creates the plsql user for chapter 9.
 *   You must run this script as SYS or
 *   SYSTEM as SYSDBA.  The script can be rerun.
 *
 * The plsql user is created using the USERS and TEMP tablespace.
 *   If these tablespaces do not exist in your environment, change
 *   the script to use an appropriate tablespaces already created.
 *
 * Modify the conn_string value below, providing your Net Service
 *   Name if not using your default.
 *
 * The value for file_loc must be the location where all your files
 *   for this example are stored.  
 *
 * text_ts and text_datafile_loc are used to create a tablespace
 *   for the Oracle Text objects.
 */

DEF conn_string = plsql/oracle@rhdb
DEF username = plsql
DEF default_ts = USERS
DEF temp_ts = TEMP
DEF file_loc = C:\files
DEF text_ts = TEXT_TS
DEF text_datafile_loc = C:\oracle\product\oradata\rhdb\text_ts01.dbf


SET FEEDBACK OFF SERVEROUTPUT ON VERIFY OFF TERMOUT OFF

SPOOL create_user.log

DECLARE
   v_count       INTEGER        := 0;
   v_statement   VARCHAR2 (500);
BEGIN

   -- Drop the user if it exists

   SELECT COUNT (1)
     INTO v_count
     FROM dba_users
    WHERE username = UPPER ('&username');

   IF v_count != 0
   THEN
      EXECUTE IMMEDIATE ('DROP USER &username CASCADE');
   END IF;

   v_count := 0;

   -- Drop the directory if it exists

   SELECT COUNT(1)
     INTO v_count
     FROM dba_directories
    WHERE UPPER(directory_name) = 'DOCUMENT_LOC';

   IF v_count != 0
   THEN
      EXECUTE IMMEDIATE ('DROP DIRECTORY DOCUMENT_LOC');
   END IF;

   -- Create the user

   v_statement :=
         'CREATE USER &username IDENTIFIED BY oracle'
      || ' DEFAULT TABLESPACE &default_ts'
      || ' TEMPORARY TABLESPACE &temp_ts'
      || ' QUOTA UNLIMITED ON &default_ts'
      || ' ACCOUNT UNLOCK';

   EXECUTE IMMEDIATE (v_statement);

   -- Grant permissions
   EXECUTE IMMEDIATE ('GRANT connect, resource TO &username');
   EXECUTE IMMEDIATE ('GRANT execute ON ctx_ddl TO &username');
   EXECUTE IMMEDIATE ('GRANT CTXAPP TO &username');
   EXECUTE IMMEDIATE ('GRANT create ANY directory, drop ANY directory TO &username');

   DBMS_OUTPUT.put_line ('	');
   DBMS_OUTPUT.put_line ('User &username created successfully');
   DBMS_OUTPUT.put_line ('	');

   -- Create tablespace for Oracle Text tables

   v_count := 0;

   SELECT COUNT (1)
     INTO v_count
     FROM dba_tablespaces
    WHERE tablespace_name = '&text_ts';

   IF v_count != 0
   THEN
      EXECUTE IMMEDIATE ('DROP TABLESPACE &text_ts INCLUDING CONTENTS');
   END IF;

   v_statement := 'CREATE TABLESPACE &text_ts DATAFILE ''&text_datafile_loc'' '
                  || 'SIZE 5M REUSE '
                  || 'AUTOEXTEND ON NEXT 10M '
                  || 'MAXSIZE 10M DEFAULT STORAGE ('
                  || 'INITIAL 512K NEXT 512K '
                  || 'MINEXTENTS 1 MAXEXTENTS 1024'
                  || 'PCTINCREASE 0 ) ';

   EXECUTE IMMEDIATE (v_statement);

EXCEPTION
   WHEN OTHERS
   THEN
      DBMS_OUTPUT.put_line (SQLERRM);
      DBMS_OUTPUT.put_line ('	');
END;
/

SET FEEDBACK ON TERMOUT ON

CREATE DIRECTORY document_loc AS '&file_loc';
GRANT READ ON DIRECTORY document_loc TO &username;

PROMPT
PROMPT Connecting to &username to create a maintenance package
PROMPT

CONN &conn_string

CREATE OR REPLACE PACKAGE clean_schema IS

   PROCEDURE tables;
   PROCEDURE ind;
 
END;
/


CREATE OR REPLACE PACKAGE BODY clean_schema AS

PROCEDURE tables
IS
   v_string VARCHAR2(50);
   v_tab_error EXCEPTION;
   PRAGMA EXCEPTION_INIT(v_tab_error, -942); 
BEGIN
   BEGIN
      v_string := 'DROP TABLE document_repository';
   EXECUTE IMMEDIATE (v_string);
   EXCEPTION
   WHEN v_tab_error
      THEN
      NULL;
   END;

   BEGIN
      v_string := 'DROP TABLE search_admin';
   EXECUTE IMMEDIATE (v_string);
   EXCEPTION
   WHEN v_tab_error
      THEN
      NULL;
   END;

   BEGIN
      v_string := 'DROP TABLE expert_gist';
   EXECUTE IMMEDIATE (v_string);
   EXCEPTION
   WHEN v_tab_error
      THEN
      NULL;
   END;

   BEGIN
      v_string := 'DROP TABLE expert_markup';
   EXECUTE IMMEDIATE (v_string);
   EXCEPTION
   WHEN v_tab_error
      THEN
      NULL;
   END;

END tables;

PROCEDURE ind
IS
   v_string VARCHAR2(50);
   v_index_error EXCEPTION;
   v_pref_error EXCEPTION;
   PRAGMA EXCEPTION_INIT(v_index_error, -1418);
   PRAGMA EXCEPTION_INIT(v_pref_error, -20000);
BEGIN
   BEGIN
      v_string := 'drop index expert_idx force';
   EXECUTE IMMEDIATE (v_string);
   EXCEPTION
   WHEN v_index_error
      THEN NULL;
   END;

   BEGIN
      v_string := 'drop index expert_title_idx force';
   EXECUTE IMMEDIATE (v_string);
   EXCEPTION
   WHEN v_index_error
      THEN NULL;
   END;

   BEGIN
      v_string := 'drop index concat_idx force';
   EXECUTE IMMEDIATE (v_string);
   EXCEPTION
   WHEN v_index_error
      THEN NULL;
   END;

   BEGIN
      v_string := 'drop index xml_idx force';
   EXECUTE IMMEDIATE (v_string);
   EXCEPTION
   WHEN v_index_error
      THEN NULL;
   END;

   BEGIN
      ctx_ddl.drop_stoplist('EXPERT_STOPLIST');
   EXCEPTION
   WHEN v_pref_error
      THEN NULL;
   END;

   BEGIN
      ctx_ddl.drop_preference('EXPERT_WORDLIST');
   EXCEPTION
   WHEN v_pref_error
      THEN NULL;
   END;

   BEGIN
      ctx_ddl.drop_preference('EXPERT_CONCAT_DATASTORE');
   EXCEPTION
   WHEN v_pref_error
      THEN NULL;
   END;

   BEGIN
      ctx_ddl.drop_preference('EXPERT_STORAGE');
   EXCEPTION
   WHEN v_pref_error
      THEN NULL;
   END;

   BEGIN
      ctx_ddl.drop_preference('EXPERT_LEXER');
   EXCEPTION
   WHEN v_pref_error
      THEN NULL;
   END;

END ind;

END;
/


SPOOL OFF
