 /*
 * dbms_datapump_objects.sql
 * Chapter 12, Expert PL/SQL Programming
 * by Ron Hardman and Michael McLaughlin
 *
 * This script creates the tables and procedures used by the
 *  dbms_datapump_export.sql and dbms_datapump_import.sql scripts.
 */

CLEAR BREAKS
CLEAR COLUMNS
CLEAR COMPUTES

TTITLE OFF


SET ECHO OFF
SET FEEDBACK ON
SET NULL '<Null>'
SET PAGESIZE 999
SET PAUSE OFF
SET TERM ON
SET TIME OFF
SET TIMING OFF
SET VERIFY OFF
SET SERVEROUTPUT ON

spool dbms_datapump_objects.log

-- Create a dummy table to export

CREATE TABLE test_datapump AS
   SELECT * FROM dba_users;
   
PROMPT How many records in the TEST_DATAPUMP table?
PROMPT
SELECT COUNT(1) FROM test_datapump;
PROMPT
   
-- Create a procedure that exports a specified table

CREATE OR REPLACE PROCEDURE export_tables (
   i_job_name VARCHAR2,
   i_dump_file_name VARCHAR2)
   
AS
  v_handle NUMBER;
  v_status VARCHAR2(500 CHAR);
  v_privs NUMBER;
  v_job_name VARCHAR2(30 CHAR) := UPPER(i_job_name);
  
  /******************************************
   * This procedure exports the TEST_DATAPUMP
   * table in the current schema, and returns
   * the completion status.
   ******************************************/
   
BEGIN

  -- Use OPEN to get the handle
  v_handle := DBMS_DATAPUMP.OPEN('EXPORT', 
                                 'TABLE',
                                 NULL, 
                                 v_job_name,
                                 NULL);
  DBMS_OUTPUT.PUT(CHR(10));
  DBMS_OUTPUT.PUT_LINE('==================================');
  DBMS_OUTPUT.PUT_LINE('==    Data Pump table export');
  DBMS_OUTPUT.PUT_LINE('==');
  DBMS_OUTPUT.PUT_LINE('== The handle is: '||v_handle);
  DBMS_OUTPUT.PUT_LINE('==================================');
  DBMS_OUTPUT.PUT(CHR(10));
  
  -- Do not discard the master table after export
  DBMS_DATAPUMP.SET_PARAMETER(v_handle, 'KEEP_MASTER', 1);
  
  -- Limit the dump to the TEST_DATAPUMP table only
  DBMS_DATAPUMP.METADATA_FILTER(v_handle, 'NAME_EXPR', 'IN (''TEST_DATAPUMP'')');
  
  -- The dump file will be placed in the OS directory
  --   associated with EXPORT_DIR
  DBMS_DATAPUMP.ADD_FILE(v_handle, i_dump_file_name, 'EXPORT_DIR');
  
  -- Start the job associated with the handle
  DBMS_DATAPUMP.START_JOB(v_handle);
  
  -- Wait until the job completes, and then get the status
  DBMS_DATAPUMP.WAIT_FOR_JOB(v_handle, v_status);
  DBMS_OUTPUT.PUT_LINE('Final status: '||v_status);
  
  -- Detach from the job
  DBMS_DATAPUMP.DETACH(v_handle);

END export_tables;
/ 

-- Import the tables into PLSQL2 schema

CREATE OR REPLACE PROCEDURE import_tables (
   i_job_name VARCHAR2,
   i_dump_file_name VARCHAR2)
   
AS
  v_handle NUMBER;
  v_status VARCHAR2(500 CHAR);
  v_privs NUMBER;
  v_job_name VARCHAR2(30 CHAR) := UPPER(i_job_name);
  
  /**************************************
   * This procedure exports all tables
   * in the current schema, and returns
   * the completion status.
   *************************************/
   
BEGIN

  -- Use OPEN to get the handle
  v_handle := DBMS_DATAPUMP.OPEN('IMPORT', 
                                 'TABLE',
                                 NULL, 
                                 v_job_name,
                                 NULL);
  DBMS_OUTPUT.PUT(CHR(10));
  DBMS_OUTPUT.PUT_LINE('==================================');
  DBMS_OUTPUT.PUT_LINE('==    Data Pump table import');
  DBMS_OUTPUT.PUT_LINE('==');
  DBMS_OUTPUT.PUT_LINE('== The handle is: '||v_handle);
  DBMS_OUTPUT.PUT_LINE('==================================');
  DBMS_OUTPUT.PUT(CHR(10));
  
  -- Specify the dump file to import
  DBMS_DATAPUMP.ADD_FILE(v_handle, i_dump_file_name, 'EXPORT_DIR');
  
  -- The log file will be placed in the OS directory
  --   associated with EXPORT_DIR
  DBMS_DATAPUMP.ADD_FILE(v_handle, 'import.log', 
                        'EXPORT_DIR', NULL,
                         DBMS_DATAPUMP.KU$_FILE_TYPE_LOG_FILE);
  
  -- Map the object owned by user PLSQL to user PLSQL2
  DBMS_DATAPUMP.METADATA_REMAP(v_handle,'REMAP_SCHEMA','PLSQL','PLSQL2');

  -- Skip the table if it already exists
  DBMS_DATAPUMP.SET_PARAMETER(v_handle,'TABLE_EXISTS_ACTION','REPLACE');
  
  -- Limit the rows imported to only those who have an
  --   ACCOUNT_STATUS value of OPEN
  DBMS_DATAPUMP.DATA_FILTER(v_handle, 'SUBQUERY', 
     'WHERE ACCOUNT_STATUS = ''OPEN''');
  
  -- Start the job associated with the handle
  DBMS_DATAPUMP.START_JOB(v_handle);
  
  -- Wait until the job completes, and then get the status
  DBMS_DATAPUMP.WAIT_FOR_JOB(v_handle, v_status);
  DBMS_OUTPUT.PUT_LINE('Final status: '||v_status);
  
  -- Detach from the job
  DBMS_DATAPUMP.DETACH(v_handle);

END import_tables;
/ 

PROMPT PLSQL Schema:  
PROMPT   Run the dbms_datapump_export.sql script, 
PROMPT   and verify the dump file was successfully created
PROMPT
PROMPT PLSQL2 Schema:
PROMPT   Run the dbms_datapump_import.sql script

spool off
